%%
%% %CopyrightBegin%
%% 
%% Copyright Ericsson AB 2002-2010. All Rights Reserved.
%% 
%% The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved online at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% %CopyrightEnd%

%%% File    : bstone.erl
%%% Purpose : Measure erlang implementations on a certain machine

-module(bstone).
-compile(export_all).

-define(TOTAL, (3000 * 1000 * 100)).   %% 300 secs
-define(BIGPROCS, 2).
-define(BIGPROC_SIZE, 50).

-define(STONEFACTOR, 31000000).   %% Factor to make the reference
                             %% implementation to make 1000 TS_ESTONES.

-record(micro,
	{function, %% The name of the function implementing the micro
	 weight,   %% How important is this in typical applications ??
	 loops = 100,%% initial data
	 tt1,      %% time to do one round
	 str}).    %% Header string

run() ->
    L = ?MODULE:macro(?MODULE:micros()),
    {Time,Stones} = sum_micros(L, 0, 0),
    [{seconds,Time/1000000.0},{stones,Stones}].

option(compiler) -> erlang:system_info(machine);
option(version) ->  erlang:system_info(version);
option(os) ->
    Name = case os:type() of
	       {_, OSName} ->
		   OSName;
	       OSName ->
		   OSName
	   end,
    case os:version() of
	{A,B,C} -> lists:concat([Name, A,".", B, ".", C]);
	Vs when list(Vs) -> lists:concat([Name, Vs])
    end.


%% In order to be able to invoke with -s flag

cmdline_run([Compiler, Version, Flags, Os, CPu, MHz]) ->
    spawn(?MODULE, run, [Compiler, Version, Flags, Os, CPu, MHz, halt]). 

tty() ->
    tty(all).
tty(all) ->
    pp(macro(micros()));
tty(MicroName) ->
    case find_micro(MicroName, micros()) of
	{ok,M} -> pp(macro([M]));
	Other -> Other
    end.

find_micro(N, [M|T]) when M#micro.function == N ->
    {ok, M};
find_micro(N, [_|T]) ->
    find_micro(N, T);
find_micro(N, []) ->
    notfound.

test() ->
    macro(determine_loop_data()).

iterate(Ofile, Times) ->
    {ok, Out} = file:open(Ofile, write),
    iterate2(Out, Times),
    file:close(Out).
iterate2(_, 0) ->
    ok;
iterate2(Out, I) ->
    pp(Out, macro(micros())),
    iterate2(Out, I-1).


micro(K, Loops) ->
    M = micro(K),
    M#micro{loops = Loops}.

micro(lists) -> 
     #micro{function = lists,
	    weight = 7, 
	    loops = 6400,
	    str = "list manipulation"};
micro(msgp) ->
    #micro{function = msgp,
	    weight = 10,
	    loops = 1515,
	    str = "small messages"};
micro(msgp_medium) ->
    #micro{function = msgp_medium,
	    weight = 14,
	    loops = 1527,
	    str = "medium messages"};
micro(msgp_huge) ->
    #micro{function = msgp_huge,
	    weight = 4,
	    loops = 52,
	    str = "huge messages"};

micro(pattern) ->
    #micro{function = pattern,
	    weight = 5,
	    loops = 1046,
	    str = "pattern matching"};

micro(trav) ->
    #micro{function = trav,
	    weight = 4,
	    loops = 2834,
	    str = "traverse"};

micro(large_dataset_work) ->
    #micro{function = large_dataset_work,
	   weight = 3,
	   loops = 1193,
	   str = "Work with large dataset"};

micro(large_local_dataset_work) ->
    #micro{function = large_local_dataset_work,
	   weight = 3,
	   loops = 1174,
	   str = "Work with large local dataset"};

micro(alloc) ->
    #micro{function = alloc,
	   weight = 2,
	   loops = 3710,
	   str = "Alloc and dealloc"};

micro(bif_dispatch) ->
    #micro{function = bif_dispatch,
	   weight = 5,
	   loops = 1623,
	   str = "Bif dispatch"};

micro(binary_h) ->
    #micro{function = binary_h,
	   weight = 4,
	   loops = 581,
	   str = "Binary handling"};
micro(ets) ->
    #micro{function = ets,
	   weight = 6,
	   loops = 342,
	   str = "ets datadictionary"};
micro(generic) ->
    #micro{function = generic,
	   weight = 9,
	   loops = 7977,
	   str = "Generic server (with timeout)"};
micro(int_arith) ->
    #micro{function = int_arith,
	   weight = 3,
	   loops = 4157,
	   str = "Small Integer arithmetics"};
micro(float_arith) ->
    #micro{function = float_arith,
	   weight = 1,
	   loops = 5526,
	   str = "Float arithmetics"};
micro(fcalls) ->
    #micro{function = fcalls,
	   weight = 5,
	   loops = 882,
	   str = "Function calls"};

micro(timer) ->
    #micro{function = timer,
	   weight = 2,
	   loops = 2312,
	   str = "Timers"};

micro(links) ->
    #micro{function = links,
	   weight = 1,
	   loops = 30,
	   str = "Links"}.



%% Return a list of micro's
micros() ->
    [
     micro(lists),
     micro(msgp),
     micro(msgp_medium),
     micro(msgp_huge),
     micro(pattern),
     micro(trav),
     micro(large_dataset_work),
     micro(large_local_dataset_work),
     micro(alloc),
     micro(bif_dispatch),
     micro(binary_h),
     micro(ets),
     micro(generic),
     micro(int_arith),
     micro(float_arith),
     micro(fcalls),
     micro(timer),
     micro(links)
    ].

macro(Ms) ->
    erlang:now(),  %% compensate for old 4.3 firsttime clock bug :-(
    statistics(reductions),
    statistics(runtime),
    lists(500),  %% fixup cache on first round
    run_micros(Ms).

run_micros([]) -> 
    io:nl(),
    [];
run_micros([H|T]) ->
    R = run_micro(H),
    [R| run_micros(T)].

run_micro(M) ->
    Pid = spawn(?MODULE, run_micro, [self(),M]),
    Res = receive {Pid, Reply} -> Reply end,
    Res.
    

run_micro(Top, M) ->
    Top ! {self(),  apply_micro(M)}.

apply_micro(M) ->
    {GC0, Words0, _} = statistics(garbage_collection),
    statistics(reductions),
    Before = erlang:now(),

    Compensate = apply_micro(M#micro.function, M#micro.loops),
    After = erlang:now(),
    {GC1, Words1, _} = statistics(garbage_collection),
    {_, Reds} = statistics(reductions),
    Elapsed = subtr(Before, After),
    MicroSecs = Elapsed - Compensate,
    [{title, M#micro.str},
     {tt1, M#micro.tt1},
     {function, M#micro.function},
     {weight_percentage, M#micro.weight},
     {loops, M#micro.loops},
     {microsecs,MicroSecs},
     {estones, (M#micro.weight * M#micro.weight * ?STONEFACTOR) div MicroSecs},
     {gcs, GC1 - GC0},
     {kilo_word_reclaimed, (Words1 - Words0) div 1000},
     {kilo_reductions, Reds div 1000},
     {gc_intensity, gci(Elapsed, GC1 - GC0, Words1 - Words0)}].


pp(Ms) ->
    pp(standard_io, Ms).

pp(Out, Ms) ->
    {Total, Stones} = sum_micros(Ms, 0, 0),
    io:format(Out, "**** Total time ~w seconds ****~n", [Total / 1000000]),
    io:format(Out, "**** ESTONES = ~w~n~n", [Stones]),
    io:format(Out, "~-31s      ~-12s  ~-10s   %    ~-10s ~n~n",
	      ["    Title", "Millis", "Estone", "Loops"]),
    pp2(Out, Ms).

sum_micros([], Tot, Stones) -> {Tot, Stones};
sum_micros([H|T], Tot, Sto) -> 
    sum_micros(T, ks(microsecs, H) + Tot, ks(estones, H) + Sto).

pp2(_,[]) ->   ok;
pp2(Out, [R|Tail]) ->
    io:format(Out, "~-35s  ~-12w    ~-10w   ~-2w    ~-10w ~n",
	      [ks(title,R), 
	       round(ks(microsecs, R) / 1000), 
	       ks(estones, R),
	       ks(weight_percentage, R),
	       ks(loops, R)]),
    pp2(Out, Tail).

ks(K, L) ->
    {value, {_, V}} = lists:keysearch(K, 1, L),
    V.

subtr(Before, After) ->
    (element(1,After)*1000000000000
     +element(2,After)*1000000+element(3,After)) -
        (element(1,Before)*1000000000000
         +element(2,Before)*1000000+element(3,Before)).

gci(Micros, Words, Gcs) ->
    ((256 * Gcs) / Micros) + (Words / Micros).

apply_micro(Name, Loops) ->
    io:format("~w(~w)~n", [Name, Loops]),
    apply(?MODULE, Name, [Loops]).

%%%%%%%%%%%% micro bench manipulating lists. %%%%%%%%%%%%%%%%%%%%%%%%%
lists(I) ->
    L1 = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    L2 = "aaaaaaaaaa",
    lists(I, L1, L2).

lists(0, _,_) ->
    0;
lists(I, L1, L2) ->
    revt(10, L1),
    appt(10, L1, L2),
    lists(I-1, L1, L2).

revt(0, _) -> 
    done;
revt(I, L) -> 
    reverse(L),
    revt(I-1, L).

reverse(L) ->
    reverse(L, []).
reverse([H|T], Ack) -> reverse(T, [H|Ack]);
reverse([], Ack) -> Ack.

append([H|T], L) ->
    [H | append(T, L)];
append([], L) ->
    L.

appt(0, L1, L2) -> ok;
appt(I, L1, L2) ->
    append(L1, L2),
    appt(I-1, L1, L2).


%%%%%%%%%%%%%%% small message passing and ctxt switching %%%%%%%
msgp(I) ->
    msgp(I, small()).

msgp(0, _) -> 
    0;
msgp(I, Msg) ->
    P1 = spawn(?MODULE, p1, [self()]),
    P2 = spawn(?MODULE, p1, [P1]),
    P3 = spawn(?MODULE, p1, [P2]),
    P4 = spawn(?MODULE, p1, [P3]),
    msgp_loop(100, P4, Msg),
    msgp(I-1, Msg).

p1(To) ->
    receive
	{From, {message, X}} ->
	    To ! {self(), {message, X}},
	    p1(To);
	stop ->
	    To ! stop,
	    exit(normal)
    end.

msgp_loop(0, P, _) ->
    P ! stop,
    receive 
	stop -> ok
    end;
msgp_loop(I, P, Msg) ->
    P ! {self(), {message, Msg}},
    receive
	{From, {message, _}} ->
	    msgp_loop(I-1, P, Msg)
    end.

%%%%%%%%%%%% large massage passing and ctxt switching %%%%%%%
msgp_medium(I) ->
        msgp_medium(I, big()).

msgp_medium(0, _) -> 
    0;
msgp_medium(I, Msg) ->
    P1 = spawn(?MODULE , p1, [self()]),
    P2 = spawn(?MODULE, p1, [P1]),
    P3 = spawn(?MODULE, p1, [P2]),
    P4 = spawn(?MODULE, p1, [P3]),
    msgp_loop(100, P4, Msg),
    msgp_medium(I-1, Msg).



%%%%%%%%%%%% huge massage passing and ctxt switching %%%%%%%
msgp_huge(I) ->
        msgp_huge(I, very_big(15)).

msgp_huge(0, _) -> 
    0;
msgp_huge(I, Msg) ->
    P1 = spawn(?MODULE , p1, [self()]),
    P4 = spawn(?MODULE, p1, [P1]),
    msgp_loop(100, P4, Msg),
    msgp_huge(I-1, Msg).

    
%%%%%% typical protocol pattern matching %%%%%%%
pattern(0) ->
    0;
pattern(I) ->
    Tail = "aaabbaaababba",
    P1 = [0, 1,2,3,4,5|Tail],
    pat_loop1(100, P1),
    pat_loop2(100, P1),
    pat_loop3(100, P1),
    pat_loop4(100, P1),
    pat_loop5(100, P1),
    pattern(I-1).

pat_loop1(0, _) -> 
    ok;
pat_loop1(I, [_,X, Y , 0 |T])  ->
    ok;
pat_loop1(I, [_, X, Y, 1| T]) ->
    ok;
pat_loop1(I, [_, X, Y, 2 | T]) -> 
    ok;
pat_loop1(I, [_, X, Y, 3 | T]) ->
    pat_loop1(I-1, [0, X,Y,3|T]).

pat_loop2(0, _) ->
    ok;
pat_loop2(I, [X, Y |Tail]) when Y bsl 1 == 0 ->
    ok;
pat_loop2(I, [X, Y | Tail]) when Y bsl 2 == 0 ->
    ok;
pat_loop2(I, [X, Y | Tail]) when Y bsl 2 == 4 ->
    pat_loop2(I-1, [X, Y |Tail]).


pat_loop3(0, _) ->
    ok;
pat_loop3(I, [{c, h} | Tail]) -> 
    ok;
pat_loop3(I, [1, 0 |T]) ->
    ok;
pat_loop3(I, [X, Y |Tail]) when binary(X), size(X) == 1 ->
    ok;
pat_loop3(I, [no, Y|Tail]) -> 
    ok;
pat_loop3(I, []) ->
    ok;
pat_loop3(I, [X,Y|T]) when X /= 0 ->
    ok;
pat_loop3(I, [2,3|T]) ->
    ok;
pat_loop3(I, [1, 2]) ->
    ok;
pat_loop3(I, [0, 1 |T]) ->
    pat_loop3(I-1, [0,1|T]).


pat_loop4(0, _) ->  ok;
pat_loop4(I, [20|T]) -> ok;
pat_loop4(I, [219|T]) -> ok;
pat_loop4(I, [18|T]) -> ok;
pat_loop4(I, [17|T]) -> ok;
pat_loop4(I, [16|T]) -> ok;
pat_loop4(I, [15|T]) -> ok;
pat_loop4(I, [14|T]) -> ok;
pat_loop4(I, [13|T]) -> ok;
pat_loop4(I, [12|T]) -> ok;
pat_loop4(I, [11|T]) -> ok;
pat_loop4(I, [10|T]) -> ok;
pat_loop4(I, [9|T]) -> ok;
pat_loop4(I, [8|T]) -> ok;
pat_loop4(I, [7|T]) -> ok;
pat_loop4(I, [6|T]) -> ok;
pat_loop4(I, [5|T]) -> ok;
pat_loop4(I, [4|T]) -> ok;
pat_loop4(I, [3|T]) -> ok;
pat_loop4(I, [1|T]) -> ok;
pat_loop4(I, [21|T]) -> ok;
pat_loop4(I, [22|T]) -> ok;
pat_loop4(I, [23|T]) -> ok;
pat_loop4(I, [24|T]) -> ok;
pat_loop4(I, [25|T]) -> ok;
pat_loop4(I, [26|T]) -> ok;
pat_loop4(I, [27|T]) -> ok;
pat_loop4(I, [0|T]) -> 
    pat_loop4(I-1, [0|T]).

pat_loop5(0, _) -> ok;
pat_loop5(I, [0, 20|T]) -> ok;
pat_loop5(I, [0, 19|T]) -> ok;
pat_loop5(I, [0, 18|T]) -> ok;
pat_loop5(I, [0, 17|T]) -> ok;
pat_loop5(I, [0, 16|T]) -> ok;
pat_loop5(I, [0, 15|T]) -> ok;
pat_loop5(I, [0, 14|T]) -> ok;
pat_loop5(I, [0, 13|T]) -> ok;
pat_loop5(I, [0, 12|T]) -> ok;
pat_loop5(I, [0, 11|T]) -> ok;
pat_loop5(I, [0, 10|T]) -> ok;
pat_loop5(I, [0, 9|T]) -> ok;
pat_loop5(I, [0, 8|T]) -> ok;
pat_loop5(I, [0, 7|T]) -> ok;
pat_loop5(I, [0, 6|T]) -> ok;
pat_loop5(I, [0, 1|T]) -> 
    pat_loop5(I-1, [0,1|T]).

%%%%%%%%%% term traversal representing simple pattern matchhing %%%
%%%%%%%%%                              + some arith
trav(I) ->
    X = very_big(10),
    trav(I, X).

trav(0, _) -> 0;
trav(I, T) ->
    do_trav(T),
    trav(I-1, T).

do_trav(T) when tuple(T) ->
    tup_trav(T, 1, 1 + size(T));
do_trav([H|T]) ->
    do_trav(H) + do_trav(T);
do_trav(X) when integer(X) -> 1;
do_trav(X) -> 0.
tup_trav(T, P, P) -> 0;
tup_trav(T, P, End) ->
    do_trav(element(P, T)) + tup_trav(T, P+1, End).

%% Working with a very large non-working data set
%% where the passive data resides in remote processes
large_dataset_work(I) ->
    {Minus, Ps} = timer:tc(?MODULE, mk_big_procs, [?BIGPROCS]),
    trav(I),
    lists(I),
    send_procs(Ps, stop),
    Minus. %% Don't count time to create the big procs.

mk_big_procs(0) -> [];
mk_big_procs(I) ->
    [ mk_big_proc()| mk_big_procs(I-1)].

mk_big_proc() ->
    P = spawn(?MODULE, big_proc, []),
    P ! {self(), running},
    receive
	{P, yes} -> P
    end.

big_proc() ->
    X = very_big(?BIGPROC_SIZE), %% creates a big heap
    Y = very_big(?BIGPROC_SIZE),
    Z = very_big(?BIGPROC_SIZE),

    receive
	{From, running} ->
	    From ! {self(), yes}
    end,
    receive
	stop ->
	    {X, Y, Z}  %% Can't be garbed away now by very (not super)
                       %% smart compiler
    end.

%% Working with a large non-working data set
%% where the data resides in the local process.
large_local_dataset_work(I) ->
    {Minus, Data} = timer:tc(?MODULE, very_big, [?BIGPROC_SIZE]),
    trav(I),
    lists(I),
    Minus.


%% Fast allocation and also deallocation that is gc test
%% Important to not let variable linger on the stack un-necessarily
alloc(0) -> 0;
alloc(I) ->
    X11 = very_big(),
    X12 = very_big(),
    X13 = very_big(),
    Z = [X14 = very_big(),
	 X15 = very_big(),
	 X16 = very_big()],
    X17 = very_big(),
    X18 = very_big(),
    X19 = very_big(),
    X20 = very_big(),
    X21 = very_big(),
    X22 = very_big(),
    X23 = very_big(),
    X24 = very_big(),
    alloc(I-1).

%% Time to call bif's
%% Lot's of element stuff which reflects the record code which
%% is becomming more and more common
bif_dispatch(0) ->
    0;
bif_dispatch(I) ->
    disp(),    disp(),    disp(),    disp(),    disp(),    disp(),
    disp(),    disp(),    disp(),    disp(),    disp(),    disp(),
    bif_dispatch(I-1).

disp() ->
    Tup = {a},
    L = [x],
    self(),self(),self(),self(),self(),self(),self(),self(),self(),
    make_ref(),
    atom_to_list(''),
    X = list_to_atom([]),
    tuple_to_list({}),
    X2 = list_to_tuple([]),
    element(1, Tup),
    element(1, Tup),
    Elem = element(1, Tup),element(1, Tup),element(1, Tup),element(1, Tup),
    element(1, Tup),element(1, Tup),element(1, Tup),element(1, Tup),
    element(1, Tup),element(1, Tup),element(1, Tup),element(1, Tup),
    element(1, Tup),element(1, Tup),element(1, Tup),element(1, Tup),
    setelement(1, Tup,k),
    setelement(1, Tup,k),
    setelement(1, Tup,k),setelement(1, Tup,k),setelement(1, Tup,k),
    setelement(1, Tup,k),setelement(1, Tup,k),setelement(1, Tup,k),
    setelement(1, Tup,k),
    setelement(1, Tup,k),
    setelement(1, Tup,k),
    setelement(1, Tup,k),
    Y = setelement(1, Tup,k),
    Date = date(), time(),
    put(a, 1),
    get(a),
    erase(a),
    hd(L),
    tl(L),
    Len = length(L),length(L),length(L),length(L),
    node(),node(),node(),node(),node(),node(),node(),node(),
    S=self(),
    node(S),node(S),node(S),
    size(Tup),
    W = whereis(code_server),whereis(code_server),
    whereis(code_server),whereis(code_server),
    whereis(code_server),whereis(code_server),
    W2 = whereis(code_server).
    
    
%% Generic server like behaviour
generic(I) ->
    register(funky, spawn(?MODULE, gserv, [funky, ?MODULE, [], []])),
    g_loop(I).

g_loop(0) ->
    exit(whereis(funky), kill),
    0;
g_loop(I) ->
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [xyz]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [xyz]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    ?MODULE:req(funky, {call, [abc]}),
    g_loop(I-1).

req(Name, Req) ->
    R = make_ref(),
    Name ! {self(), R, Req},
    receive
	{Name, R, Reply} -> Reply
    after 2000 ->
	    exit(timeout)
    end.

gserv(Name, Mod, State, Debug) ->
    receive
	{From, Ref, {call, Req}} when Debug == [] ->
	    case catch apply(Mod, handle_call, [From, State, Req]) of
		{reply, Reply, State2} ->
		    From ! {Name, Ref, Reply},
		    gserv(Name, Mod, State2, Debug);
		{noreply, State2} ->
		    gserv(Name, Mod, State2, Debug);
		{'EXIT', Reason} ->
		    exit(Reason)
	    end;
	{From, Ref, Req} when Debug /= [] ->
	    exit(nodebug)
    end.

handle_call(From, State, [xyz]) ->
    R = atom_to_list(xyz),
    {reply, R, []};
handle_call(From, State, [abc]) ->
    R = 1 + 3,
    {reply, R, [R | State]}.

		    

%% Binary handling, creating, manipulating and sending binaries
binary_h(I) ->
    Before = erlang:now(),
    P = spawn(?MODULE, echo, [self()]),
    B = list_to_binary(lists:duplicate(2000, 5)),
    After = erlang:now(),
    Compensate = subtr(Before, After),
    binary_h_2(I, P, B),
    Compensate.
    
binary_h_2(0, P, B) ->
    exit(P, kill);
binary_h_2(I, P, B) ->
    echo_loop(P, 20, B),
    split_loop(B, {abc,1,2222,self(),"ancnd"}, 100),
    binary_h_2(I-1, P, B).

split_loop(B, _, 0) -> 
    ok;
split_loop(B, Term, I) ->
    {X, Y} = split_binary(B, I),
    size(X),
    binary_to_list(Y, 1, 2),
    binary_to_term(term_to_binary(Term)),
    split_loop(B, Term, I-1).
    

echo_loop(P, 0, B) -> 
    k;
echo_loop(P, I, B) ->
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    receive _ -> ok end,
    echo_loop(P, I-1, B).
    

ets(0) -> 
    0;
ets(I) ->
    T1 = ets:new(a, [set]),
    T2 = ets:new(c, [bag, private]),
    L = [T1, T2],
    run_tabs(L, L, 1),
    ets:delete(T1),
    ets:delete(T2),
    ets(I-1).

run_tabs(_, _, 0) ->
    ok;
run_tabs([], L, I) ->
    run_tabs(L, L, I-1);
run_tabs([Tab|Tail], L, I) ->
    Begin = I * 20,
    End = (I+1) * 20,
    run_tab(Tab, Begin, End, I),
    run_tabs(Tail, L, I).

run_tab(Tab, X, X, _) ->
    ok;
run_tab(Tab, Beg, End, J) ->
    ets:insert(Tab, {Beg, J}),
    ets:insert(Tab, {J, Beg}),
    ets:insert(Tab, {{foo,Beg}, J}),
    ets:insert(Tab, {{foo, J}, Beg}),
    ets:delete(Tab, haha),
    ets:match_delete(Tab, {k, j}),
    ets:match(Tab, {Beg, '$1'}),
    ets:match(Tab, {'$1', J}),
    ets:delete(Tab, Beg),
    K = ets:first(Tab),
    K2 = ets:next(Tab, K),
    run_tab(Tab, Beg+1, End, J).
    
    
%%%% Integer arith %%%%%
int_arith(0) -> 
    0;
int_arith(I) ->
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
    do_arith(I) +
	66,
    int_arith(I-1).

do_arith(I) ->    
    do_arith2(I) -
    do_arith2(I) -
    do_arith2(I) -
    do_arith2(I) -
    do_arith2(I) -
    do_arith2(I) -
    do_arith2(I) -
	99.

do_arith2(I) ->
    X = 23,
    Y = 789 + I,
    Z = I + 1,
    U = (X bsl 1 bsr I) * X div 2 bsr 4,
    U1 = Z + Z + Z + Z + X bsl 4 * 2 bsl 2,
    Z - U + U1 div 2.

    
%%%% Float arith %%%%%
float_arith(0) -> 
    0;
float_arith(I) ->
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
    f_do_arith(I) +
	66,
    float_arith(I-1).

f_do_arith(I) ->    
    X = 23.4,
    Y = 789.99 + I,
    Z = I + 1.88,
    U = (X * 1 / I) * X / 2 * 4,
    U1 = Z + Z + Z + Z + X * 4 * 2 / 2,
    Z - U + U1 / 2.

%%%% time to do various function calls
fcalls(0) -> 
    0;
fcalls(I) ->
    local0(400),
    remote0(400),
    app0(400),
    local1(400),
    remote1(400),
    app1(400),
    fcalls(I-1).


local0(0) -> 0;
local0(N) ->
    local0(N-1).

local1(0) -> 0;
local1(N) ->
    1+local1(N-1).

remote0(0) -> 0;
remote0(N) ->
    ?MODULE:remote0(N-1).

remote1(0) -> 0;
remote1(N) ->
    1+?MODULE:remote1(N-1).

app0(0) -> 0;
app0(N) ->
    apply(?MODULE, app0, [N-1]).

app1(0) -> 0;
app1(N) ->
    1 + apply(?MODULE, app1, [N-1]).

%%%%%% jog the time queue implementation
timer(I) ->
    L = [50, 50, 50, 100, 1000, 3000, 8000, 50000, 100000],
    timer(I, L).

timer(0, _) -> 0;
timer(N, L) ->
    send_self(100),
    recv(100,L, L),
    timer(N-1).

recv(0, _, _) ->
    ok;
recv(N, [], L) ->
    recv(N, L, L);
recv(N, [Timeout|Tail], L) ->
    receive
        hi_dude ->
            recv(N-1, Tail, L)
    after Timeout ->
            io:format("XXXXX this wasn't supposed to happen???~n", []),
            ok
    end.

send_self(0) ->
    ok;
send_self(N) ->
    self() ! hi_dude,
    send_self(N-1).


%%%%%%%%%%%% managing many links %%%%%
links(I) ->
    L = mk_link_procs(100),
    send_procs(L, {procs, L, I}),
    wait_for_pids(L),
    0.

mk_link_procs(0) -> 
    [];
mk_link_procs(I) ->
    [spawn_link(?MODULE, lproc, [self()]) | mk_link_procs(I-1)].


lproc(Top) ->
    process_flag(trap_exit,true),
    receive
	{procs, Procs, I} ->
	    Top ! {self(), lproc(Procs, Procs, link, I)}
    end.

lproc(_, _, _, 0) ->
    done;
lproc([], Procs, link, I) ->
    lproc(Procs, Procs, unlink, I-1);
lproc([], Procs, unlink, I) ->
    lproc(Procs, Procs, link, I-1);
lproc([Pid|Tail], Procs, unlink, I) ->
    unlink(Pid),
    lproc(Tail, Procs, unlink, I);
lproc([Pid|Tail], Procs, link, I) ->
    link(Pid),
    lproc(Tail, Procs, unlink, I).



%%%%%%%%%%% various utility functions %%%%%%%

echo(Pid) ->
    receive
	X -> Pid ! X,
	     echo(Pid)
    end.

very_big() -> 
    very_big(2).
very_big(0) -> [];
very_big(I) ->
    {1,2,3,a,v,f,r,t,y,u,self(), self(), self(), 
     "22222222222222222", {{"234", self()}}, 
     [[very_big(I-1)]]}.
 
big() ->
    {self(), funky_stuff, baby, {1, [123, true,[]], "abcdef"}}.

small() -> {self(), true}.    
    
%% Wait for a list of children to respond    
wait_for_pids([]) -> 
    ok;
wait_for_pids([P|Tail]) ->
    receive 
	{P, Res} -> wait_for_pids(Tail)
    end.

send_procs([P|Tail], Msg) -> P ! Msg, send_procs(Tail, Msg);
send_procs([], _) -> ok.
			     
%% Time to perform one subtraction + one function call
loopt() ->
    Before = erlang:now(),
    loopt(10000),
    After = erlang:now(),
    subtr(Before, After) / 10000.

loopt(0) -> ok;
loopt(I) -> loopt(I-1).


%% This function is run be me only (me being klacke) when
%% I've determined how important the different micros are.
%% The purpose is to determine the loops field for each micro
%% in the final bench.

determine_loop_data() ->
    determine_loop_data(1).
determine_loop_data(I) ->
    Mics = micros(),
    ppld([Mics | determine_loop_data(I, Mics)]).
determine_loop_data(0, Ms) ->
    [];
determine_loop_data(I, Ms) ->
    io:format("Determining loop counts \n",[]),
    check_sum_micros(Ms, 0),
    Tmp = tt1(Ms),
    Ms2 = set_loops(Tmp),
    [Ms2 | determine_loop_data(I-1, Ms2)]. %% fixpoint iterate
    
ppld(Mss) ->
    ppld(micros(), Mss).
ppld([], _) ->
    ok;
ppld([M| Tail], Mss) ->
    Str = M#micro.str,
    Loops = find_loops(M#micro.function, Mss),
    io:format("~-20s  Adjusted Mean: ~-5w   ~w~n",
	      [Str, mean(Loops), Loops]),
    ppld(Tail, Mss).

mean(L0) ->
    L1 = lists:sort(L0),  %% romove the 2 extremes
    L = lists:delete(lists:last(L1), tl(L1)),
    round(sum(L) / length(L)).

sum([]) -> 0;
sum([H|T]) -> H + sum(T).

find_loops(Fn, [MicroList | Tail]) ->
    {ok, M} = find_micro(Fn, MicroList),
    [M#micro.loops | find_loops(Fn, Tail)];
find_loops(_, []) ->
    [].

    

%% We want the macro to take ?TOTAL microsecs
%% This loop runs micro M, M#micro.loops times,
%% and the figures out how what the proper loops value
%% should be in order to spend wight percent of tge ?TOTAL time
%% doing micro M (on the reference implementation).

set_loops([M|Tail]) ->
    TT1 = M#micro.tt1,
    MicroTime = (M#micro.weight * ?TOTAL * 0.01),
    Loops = round(MicroTime / TT1),
    if
	Loops == M#micro.loops -> 
	    io:format("~s is the same~n", [M#micro.str]);
	true ->
	    io:format("Adjusting ~s loop val from ~w to ~w~n",
		      [M#micro.str, M#micro.loops, Loops])
    end,

    [M#micro{loops = Loops} | set_loops(Tail)];
set_loops([]) ->
    [].

check_sum_micros([], 100) -> ok;
check_sum_micros([], I) -> io:format("Micros don't sum up ~w~n", [I]);
check_sum_micros([H|T], Ack) -> check_sum_micros(T, Ack+H#micro.weight).

tt1([M|Tail]) ->
    X = run_micro(M#micro{loops = 500}),
    {value, {_, Ms}} = lists:keysearch(microsecs, 1, X),
    New = M#micro{tt1 = Ms / 500},
    [New | tt1(Tail)];
tt1([]) -> 
    io:nl(),
    [].

%[mk_fname([Compiler, Version, Flags, Os, CPu, MHz])]),
mk_fname([Compiler, Version, Flags, Os, Cpu, Mhz]) ->
    FlagStr=mk_flagstring(Flags),
    lists:flatten(
      lists:concat(
	[Compiler, "-", Version, FlagStr, "-", Os,"-", Cpu,"-", Mhz])).

mk_flagstring(List) ->
    mk_flagstring(lists:reverse(List), []).
mk_flagstring([L|Ls], Acc) ->
    mk_flagstring(Ls, [lists:concat([Acc,"-",L])]);
mk_flagstring([], Acc) ->
    Acc.
